# cython: language_level=3
# distutils: language = c++
# pylint: disable=no-member, used-before-assignment, no-name-in-module, import-error
"""
Helper Cython functions for discrete representation defined using the pure Python syntax.

This module needs to be compiled!
"""
from typing import List, Set, Tuple

import cython
import cython.cimports.libc.math as math_c
import numpy as np
from cython.cimports.libcpp import bool as bool_C
from cython.cimports.libcpp.stack import stack
from cython.cimports.libcpp.vector import vector
from numpy.typing import NDArray

# CUSTOM TYPES
_Point3D = Tuple[float, float, float]
_Triangle3D = Tuple[_Point3D, _Point3D, _Point3D]

_Point2D = Tuple[float, float]
_Segment2D = Tuple[_Point2D, _Point2D]


# PYTHON FUNCTIONS


def triangles_to_voxel_matrix(
    triangles: List[_Triangle3D],
    voxel_size: float,
) -> Tuple[NDArray[np.bool_], Tuple[float, float, float]]:
    """
    Helper function to compute the voxel matrix of all the voxels intersecting with a given list of triangles.

    :param triangles: The triangles to compute the intersecting voxels.
    :type triangles: list[tuple[tuple[float, float, float], tuple[float, float, float], tuple[float, float, float]]]
    :param voxel_size: The voxel edges size.
    :type voxel_size: float

    :return: The voxel matrix and the origin center of the matrix.
    :rtype: tuple[np.ndarray[np.bool_, np.ndim == 3], tuple[float, float, float]]
    """
    # compute the size of the matrix and min matrix origin center
    min_point, max_point = _triangles_min_max_points(triangles)
    shape = (
        int(max_point[0] // voxel_size + 1) - int(min_point[0] // voxel_size) + 2,
        int(max_point[1] // voxel_size + 1) - int(min_point[1] // voxel_size) + 2,
        int(max_point[2] // voxel_size + 1) - int(min_point[2] // voxel_size) + 2,
    )
    matrix = np.zeros(shape, dtype=np.bool_)
    matrix_origin_center = (
        round((min_point[0] // voxel_size - 0.5) * voxel_size, 9),
        round((min_point[1] // voxel_size - 0.5) * voxel_size, 9),
        round((min_point[2] // voxel_size - 0.5) * voxel_size, 9),
    )

    # compute the intersecting voxel
    return (
        np.asarray(
            _triangles_to_voxel_matrix(
                np.array(triangles),
                len(triangles),
                voxel_size,
                matrix,
                matrix_origin_center,
            ),
            np.bool_,
        ),
        matrix_origin_center,
    )


def flood_fill_matrix_2d(matrix: NDArray[np.bool_], start: Tuple[int, int], fill_with: bool) -> NDArray[np.bool_]:
    """
    Perform 2D flood fill on the given matrix starting from the specified position.

    :param matrix: The matrix to perform flood fill on.
    :type matrix: np.ndarray[np.bool_, np.ndim == 2]
    :param start: The starting position for flood fill.
    :type start: Tuple[int, int]
    :param fill_with: The value to fill the connected component with.
    :type fill_with: bool

    :return: The matrix after performing flood fill.
    :rtype: np.ndarray[np.bool_, np.ndim == 2]
    """
    if not (0 <= start[0] < matrix.shape[0] and 0 <= start[1] < matrix.shape[1]):
        raise IndexError("Start index is outside the matrix.")

    return np.asarray(
        _flood_fill_matrix_2d(
            matrix.astype(np.bool_),
            (start[0], start[1]),
            fill_with,
            (matrix.shape[0], matrix.shape[1]),
        ),
        dtype=np.bool_,
    )


def flood_fill_matrix_3d(matrix: NDArray[np.bool_], start: Tuple[int, int, int], fill_with: bool) -> NDArray[np.bool_]:
    """
    Perform 3D flood fill on the given matrix starting from the specified position.

    :param matrix: The matrix to perform flood fill on.
    :type matrix: np.ndarray[np.bool_, np.ndim == 3]
    :param start: The starting position for flood fill.
    :type start: tuple[int, int, int]
    :param fill_with: The value to fill the connected component with.
    :type fill_with: bool

    :return: The matrix after performing flood fill.
    :rtype: np.ndarray[np.bool_, np.ndim == 3]
    """
    if not (0 <= start[0] < matrix.shape[0] and 0 <= start[1] < matrix.shape[1] and 0 <= start[2] < matrix.shape[2]):
        raise IndexError("Start index is outside the matrix.")

    return np.asarray(
        _flood_fill_matrix_3d(
            matrix.astype(np.bool_),
            (start[0], start[1], start[2]),
            fill_with,
            (matrix.shape[0], matrix.shape[1], matrix.shape[2]),
        ),
        dtype=np.bool_,
    )


def line_segments_to_pixels(line_segments: List[_Segment2D], pixel_size: float) -> Set[_Point2D]:
    """
    Convert a list of line segments to a set of pixel coordinates.

    This function takes a list of line segments, each defined by two endpoints, and converts them into a set of pixel
    coordinates based on the specified pixel size. The pixel coordinates represent the pixels that the line segments
    intersect with.

    :param line_segments: List of line segments to convert.
    :type line_segments: list[tuple[tuple[float, float], tuple[float, float]]]
    :param pixel_size: The size of each pixel.
    :type pixel_size: float

    :return: A set of pixel coordinates representing the intersection points of the line segments.
    :rtype: set[tuple[float, float]]
    """
    return set(_line_segments_to_pixels(line_segments, pixel_size))


def triangle_intersects_voxel(
    triangle: _Triangle3D, voxel_center: _Point3D, voxel_extents: Tuple[float, float, float]
) -> bool:
    """
    Helper function to compute if there is an intersection between a 3D triangle and a voxel.

    This function uses the "Separating Axis Theorem".

    :param triangle: The triangle to check if it intersects with the voxel.
    :type: triangle: tuple[tuple[float, float, float], tuple[float, float, float], tuple[float, float, float]]
    :param voxel_center: The center point of the voxel.
    :type voxel_center: tuple[float, float, float]
    :param voxel_extents: The extents of the voxel in each direction (half-size of the voxel size).
    :type voxel_extents: list[float, float, float]

    :return: True if there is an intersection, False otherwise.
    :rtype: bool
    """
    return _triangle_intersects_voxel(triangle, voxel_center, voxel_extents)


def voxel_triangular_faces(voxel_center: _Point3D, voxel_size: float) -> List[_Triangle3D]:
    """
    Helper function to compute the 12 triangular faces that compose a voxel, for visualization.

    :param voxel_center: The voxel center point.
    :type voxel_center: tuple[float, float, float]
    :param voxel_size: The voxel edges size.
    :type voxel_size: float

    :return: The 12 triangles representing the 6 faces of the given voxel.
    """
    return _voxel_triangular_faces(voxel_center[0], voxel_center[1], voxel_center[2], voxel_size)


# CYTHON FUNCTIONS


@cython.cfunc
@cython.cdivision(True)
@cython.exceptval(check=False)
def _round_to_digits(num: cython.double, digits: cython.int) -> cython.double:
    """Round the given number to the specified number of digits after the decimal point."""
    multiplier: cython.double = math_c.pow(10.0, digits)
    return math_c.round(num * multiplier) / multiplier


@cython.cfunc
@cython.cdivision(True)
@cython.exceptval(check=False)
def _round_point_3d_to_digits(
    point_3d: Tuple[cython.double, cython.double, cython.double], digits: cython.int
) -> Tuple[cython.double, cython.double, cython.double]:
    """Round the given point to the specified number of digits after the decimal point."""
    return (
        _round_to_digits(point_3d[0], digits),
        _round_to_digits(point_3d[1], digits),
        _round_to_digits(point_3d[2], digits),
    )


@cython.cfunc
@cython.exceptval(check=False)
def _is_integer(value: cython.double) -> bool_C:
    """Check if the given value is an integer (has no fractional part)."""
    return cython.cast(cython.int, value) == value


@cython.cfunc
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.exceptval(check=False)
def _triangle_intersects_voxel(
    triangle: Tuple[
        Tuple[cython.double, cython.double, cython.double],
        Tuple[cython.double, cython.double, cython.double],
        Tuple[cython.double, cython.double, cython.double],
    ],
    voxel_center: Tuple[cython.double, cython.double, cython.double],
    voxel_extents: Tuple[cython.double, cython.double, cython.double],
) -> bool_C:
    """Check if a 3D triangle intersects with a voxel defined by its center and extents."""
    # Ported from https://gist.github.com/zvonicek/fe73ba9903f49d57314cf7e8e0f05dcf
    # pylint: disable=too-many-locals, too-many-return-statements, too-many-branches, too-many-statements, invalid-name

    v0: cython.double[3]
    v1: cython.double[3]
    v2: cython.double[3]
    f0: cython.double[3]
    f1: cython.double[3]
    f2: cython.double[3]
    plane_normal: cython.double[3]
    plane_distance: cython.double
    radius: cython.double
    a00: cython.double[3]
    a01: cython.double[3]
    a02: cython.double[3]
    a10: cython.double[3]
    a11: cython.double[3]
    a12: cython.double[3]
    a20: cython.double[3]
    a21: cython.double[3]
    a22: cython.double[3]

    # Translate triangle as conceptually moving AABB to origin
    v0[0] = triangle[0][0] - voxel_center[0]
    v0[1] = triangle[0][1] - voxel_center[1]
    v0[2] = triangle[0][2] - voxel_center[2]

    v1[0] = triangle[1][0] - voxel_center[0]
    v1[1] = triangle[1][1] - voxel_center[1]
    v1[2] = triangle[1][2] - voxel_center[2]

    v2[0] = triangle[2][0] - voxel_center[0]
    v2[1] = triangle[2][1] - voxel_center[1]
    v2[2] = triangle[2][2] - voxel_center[2]

    # Compute edge vectors for triangle
    f0[0] = triangle[1][0] - triangle[0][0]
    f0[1] = triangle[1][1] - triangle[0][1]
    f0[2] = triangle[1][2] - triangle[0][2]

    f1[0] = triangle[2][0] - triangle[1][0]
    f1[1] = triangle[2][1] - triangle[1][1]
    f1[2] = triangle[2][2] - triangle[1][2]

    f2[0] = triangle[0][0] - triangle[2][0]
    f2[1] = triangle[0][1] - triangle[2][1]
    f2[2] = triangle[0][2] - triangle[2][2]

    # REGION TEST THE THREE AXES CORRESPONDING TO THE FACE NORMALS OF AABB B (CATEGORY 1)

    # Exit if...
    # ... [-extents.X, extents.X] and [min(v0.X,v1.X,v2.X), max(v0.X,v1.X,v2.X)] do not overlap
    if max(v0[0], v1[0], v2[0]) < -voxel_extents[0] or min(v0[0], v1[0], v2[0]) > voxel_extents[0]:
        return False

    # ... [-extents.Y, extents.Y] and [min(v0.Y,v1.Y,v2.Y), max(v0.Y,v1.Y,v2.Y)] do not overlap
    if max(v0[1], v1[1], v2[1]) < -voxel_extents[1] or min(v0[1], v1[1], v2[1]) > voxel_extents[1]:
        return False

    # ... [-extents.Z, extents.Z] and [min(v0.Z,v1.Z,v2.Z), max(v0.Z,v1.Z,v2.Z)] do not overlap
    if max(v0[2], v1[2], v2[2]) < -voxel_extents[2] or min(v0[2], v1[2], v2[2]) > voxel_extents[2]:
        return False

    # ENDREGION

    # REGION TEST SEPARATING AXIS CORRESPONDING TO TRIANGLE FACE NORMAL (CATEGORY 2)

    plane_normal[0] = f0[1] * f1[2] - f0[2] * f1[1]
    plane_normal[1] = f0[2] * f1[0] - f0[0] * f1[2]
    plane_normal[2] = f0[0] * f1[1] - f0[1] * f1[0]

    plane_distance = math_c.fabs(plane_normal[0] * v0[0] + plane_normal[1] * v0[1] + plane_normal[2] * v0[2])

    # Compute the projection interval radius of b onto L(t) = b.c + t * p.n
    radius = (
        voxel_extents[0] * math_c.fabs(plane_normal[0])
        + voxel_extents[1] * math_c.fabs(plane_normal[1])
        + voxel_extents[2] * math_c.fabs(plane_normal[2])
    )

    # Intersection occurs when plane distance falls within [-radius,+radius] interval
    if plane_distance > radius:
        return False

    # ENDREGION

    # REGION TEST AXES a00..a22 (CATEGORY 3)

    # Test axis a00
    a00[0] = 0
    a00[1] = -f0[2]
    a00[2] = f0[1]
    if _calculate_axis_values(v0, v1, v2, a00, f0, voxel_extents):
        return False

    # Test axis a01
    a01[0] = 0
    a01[1] = -f1[2]
    a01[2] = f1[1]
    if _calculate_axis_values(v0, v1, v2, a01, f1, voxel_extents):
        return False

    # Test axis a02
    a02[0] = 0
    a02[1] = -f2[2]
    a02[2] = f2[1]
    if _calculate_axis_values(v0, v1, v2, a02, f2, voxel_extents):
        return False

    # Test axis a10
    a10[0] = f0[2]
    a10[1] = 0
    a10[2] = -f0[0]
    if _calculate_axis_values(v0, v1, v2, a10, f0, voxel_extents):
        return False

    # Test axis a11
    a11[0] = f1[2]
    a11[1] = 0
    a11[2] = -f1[0]
    if _calculate_axis_values(v0, v1, v2, a11, f1, voxel_extents):
        return False

    # Test axis a12
    a12[0] = f2[2]
    a12[1] = 0
    a12[2] = -f2[0]
    if _calculate_axis_values(v0, v1, v2, a12, f2, voxel_extents):
        return False

    # Test axis a20
    a20[0] = -f0[1]
    a20[1] = f0[0]
    a20[2] = 0
    if _calculate_axis_values(v0, v1, v2, a20, f0, voxel_extents):
        return False

    # Test axis a21
    a21[0] = -f1[1]
    a21[1] = f1[0]
    a21[2] = 0
    if _calculate_axis_values(v0, v1, v2, a21, f1, voxel_extents):
        return False

    # Test axis a22
    a22[0] = -f2[1]
    a22[1] = f2[0]
    a22[2] = 0
    if _calculate_axis_values(v0, v1, v2, a22, f2, voxel_extents):
        return False

    # ENDREGION

    return True


@cython.cfunc
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.exceptval(check=False)
def _calculate_axis_values(
    v0: cython.double[3],
    v1: cython.double[3],
    v2: cython.double[3],
    ax: cython.double[3],
    f: cython.double[3],
    voxel_extents: Tuple[cython.double, cython.double, cython.double],
) -> bool_C:
    """Calculate axis values used in triangle intersection tests with an axis-aligned box."""
    # pylint: disable=invalid-name

    p0 = v0[0] * ax[0] + v0[1] * ax[1] + v0[2] * ax[2]
    p1 = v1[0] * ax[0] + v1[1] * ax[1] + v1[2] * ax[2]
    p2 = v2[0] * ax[0] + v2[1] * ax[1] + v2[2] * ax[2]
    radius = (
        voxel_extents[0] * math_c.fabs(f[2])
        + voxel_extents[1] * math_c.fabs(f[0])
        + voxel_extents[2] * math_c.fabs(f[1])
    )

    return max(-max(p0, p1, p2), min(p0, p1, p2)) > radius


@cython.cfunc
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
@cython.exceptval(check=False)
def _aabb_intersecting_boxes(
    min_point: Tuple[cython.double, cython.double, cython.double],
    max_point: Tuple[cython.double, cython.double, cython.double],
    voxel_size: cython.double,
) -> vector[Tuple[cython.double, cython.double, cython.double]]:
    """Calculate the voxel centers that intersects with a given axis-aligned boxes."""

    x_start: cython.int
    x_end: cython.int
    y_start: cython.int
    y_end: cython.int
    z_start: cython.int
    z_end: cython.int
    x: cython.int
    y: cython.int
    z: cython.int
    num_centers: cython.int

    x_start = cython.cast(cython.int, (min_point[0] / voxel_size) - 1)
    x_end = cython.cast(cython.int, (max_point[0] / voxel_size) + 1)
    y_start = cython.cast(cython.int, (min_point[1] / voxel_size) - 1)
    y_end = cython.cast(cython.int, (max_point[1] / voxel_size) + 1)
    z_start = cython.cast(cython.int, (min_point[2] / voxel_size) - 1)
    z_end = cython.cast(cython.int, (max_point[2] / voxel_size) + 1)

    num_centers = (x_end - x_start) * (y_end - y_start) * (z_end - z_start)
    centers: vector[Tuple[cython.double, cython.double, cython.double]]
    centers.resize(num_centers)

    num_centers = 0
    for x in range(x_start, x_end):
        for y in range(y_start, y_end):
            for z in range(z_start, z_end):
                centers[num_centers] = (
                    _round_to_digits((x + 0.5) * voxel_size, 9),
                    _round_to_digits((y + 0.5) * voxel_size, 9),
                    _round_to_digits((z + 0.5) * voxel_size, 9),
                )
                num_centers += 1

    return centers


@cython.cfunc
@cython.boundscheck(False)
@cython.wraparound(False)
def _flood_fill_matrix_2d(
    matrix: bool_C[:, :],
    start: Tuple[cython.int, cython.int],
    fill_with: bool_C,
    shape: Tuple[cython.int, cython.int],
) -> bool_C[:, :]:
    """Apply a flood fill algorithm to a 2D boolean matrix from a given starting point."""

    dx: cython.int[4] = [0, 0, -1, 1]
    dy: cython.int[4] = [-1, 1, 0, 0]
    new_x: cython.int
    new_y: cython.int
    x: cython.int
    y: cython.int

    old_value: cython.int = matrix[start[0], start[1]]

    if old_value == fill_with:
        return matrix

    fill_stack: stack[Tuple[cython.int, cython.int]]
    fill_stack.push((start[0], start[1]))

    while not fill_stack.empty():
        x, y = fill_stack.top()
        fill_stack.pop()
        matrix[x, y] = fill_with

        for i in range(4):
            new_x, new_y = x + dx[i], y + dy[i]

            if 0 <= new_x < shape[0] and 0 <= new_y < shape[1] and matrix[new_x, new_y] == old_value:
                fill_stack.push((new_x, new_y))

    return matrix


@cython.cfunc
@cython.boundscheck(False)
@cython.wraparound(False)
def _flood_fill_matrix_3d(
    matrix: bool_C[:, :, :],
    start: Tuple[cython.int, cython.int, cython.int],
    fill_with: bool_C,
    shape: Tuple[cython.int, cython.int, cython.int],
) -> bool_C[:, :, :]:
    """Apply a flood fill algorithm to a 3D boolean matrix from a given starting point."""
    # pylint: disable=too-many-locals

    dx: cython.int[6] = [0, 0, -1, 1, 0, 0]
    dy: cython.int[6] = [-1, 1, 0, 0, 0, 0]
    dz: cython.int[6] = [0, 0, 0, 0, -1, 1]
    new_x: cython.int
    new_y: cython.int
    new_z: cython.int
    x: cython.int
    y: cython.int
    z: cython.int

    old_value: cython.int = matrix[start[0], start[1], start[2]]

    if old_value == fill_with:
        return matrix

    fill_stack: stack[Tuple[cython.int, cython.int, cython.int]]
    fill_stack.push((start[0], start[1], start[2]))

    while not fill_stack.empty():
        x, y, z = fill_stack.top()
        fill_stack.pop()
        matrix[x, y, z] = fill_with

        for i in range(6):
            new_x, new_y, new_z = x + dx[i], y + dy[i], z + dz[i]

            if (
                0 <= x + dx[i] < shape[0]
                and 0 <= y + dy[i] < shape[1]
                and 0 <= z + dz[i] < shape[2]
                and matrix[new_x, new_y, new_z] == old_value
            ):
                fill_stack.push((new_x, new_y, new_z))

    return matrix


@cython.cfunc
@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.exceptval(check=False)
def _line_segment_intersects_pixel(
    x1: cython.double,
    y1: cython.double,
    x2: cython.double,
    y2: cython.double,
    pixel_center_x: cython.double,
    pixel_center_y: cython.double,
    pixel_size: cython.double,
) -> bool_C:
    """Check if a line segment intersects with a pixel defined by its center and size."""
    # pylint: disable=too-many-arguments, too-many-locals

    # Determine the coordinates of lower-left and upper-right of rectangle
    xmin, xmax = pixel_center_x - pixel_size / 2, pixel_center_x + pixel_size / 2
    ymin, ymax = pixel_center_y - pixel_size / 2, pixel_center_y + pixel_size / 2

    # Compute the line equation for a point
    line_eq1 = _round_to_digits((y2 - y1) * xmin + (x1 - x2) * ymin + (x2 * y1 - x1 * y2), 9)
    line_eq2 = _round_to_digits((y2 - y1) * xmin + (x1 - x2) * ymax + (x2 * y1 - x1 * y2), 9)
    line_eq3 = _round_to_digits((y2 - y1) * xmax + (x1 - x2) * ymin + (x2 * y1 - x1 * y2), 9)
    line_eq4 = _round_to_digits((y2 - y1) * xmax + (x1 - x2) * ymax + (x2 * y1 - x1 * y2), 9)

    # Check if all corners are on the same side of the line
    miss: bool_C = (
        (line_eq1 >= 0 and line_eq2 >= 0 and line_eq3 >= 0 and line_eq4 >= 0)
        or (line_eq1 < 0 and line_eq2 < 0 and line_eq3 < 0 and line_eq4 < 0)
    ) and (
        (line_eq1 > 0 and line_eq2 > 0 and line_eq3 > 0 and line_eq4 > 0)
        or (line_eq1 <= 0 and line_eq2 <= 0 and line_eq3 <= 0 and line_eq4 <= 0)
    )

    # Does it miss based on the shadow intersection test?
    shadow_miss: bool_C = (
        (x1 > xmax and x2 > xmax) or (x1 < xmin and x2 < xmin) or (y1 > ymax and y2 > ymax) or (y1 < ymin and y2 < ymin)
    )

    # A hit is if it doesn't miss on both tests!
    return not (miss or shadow_miss)


@cython.cfunc
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def _line_segments_to_pixels(
    line_segments: vector[Tuple[Tuple[cython.double, cython.double], Tuple[cython.double, cython.double]]],
    pixel_size: cython.double,
) -> vector[Tuple[cython.double, cython.double]]:
    """Convert line segments to a list of pixel centers they intersect with."""
    # pylint: disable=too-many-locals

    pixel_centers: vector[Tuple[cython.double, cython.double]]

    for i in range(line_segments.size()):
        x1: cython.double = line_segments[i][0][0]
        y1: cython.double = line_segments[i][0][1]
        x2: cython.double = line_segments[i][1][0]
        y2: cython.double = line_segments[i][1][1]

        # Calculate the bounding box of the line segment
        xmin = min(x1, x2)
        ymin = min(y1, y2)
        xmax = max(x1, x2)
        ymax = max(y1, y2)

        # Calculate the indices of the box that intersect with the bounding box of the line segment
        x_start = cython.cast(cython.int, (xmin / pixel_size) - 2)
        x_end = cython.cast(cython.int, (xmax / pixel_size) + 2)
        y_start = cython.cast(cython.int, (ymin / pixel_size) - 2)
        y_end = cython.cast(cython.int, (ymax / pixel_size) + 2)

        # Create a list of the centers of all the intersecting voxels
        for x in range(x_start, x_end):
            for y in range(y_start, y_end):
                x_coord: cython.double = (cython.cast(cython.double, x) + 0.5) * pixel_size
                y_coord: cython.double = (cython.cast(cython.double, y) + 0.5) * pixel_size
                center: Tuple[cython.double, cython.double] = (
                    _round_to_digits(x_coord, 9),
                    _round_to_digits(y_coord, 9),
                )

                if _line_segment_intersects_pixel(x1, y1, x2, y2, center[0], center[1], pixel_size):
                    pixel_centers.push_back(center)

    return pixel_centers


@cython.cfunc
@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def _triangle_2d_to_pixels(
    triangle_2d: Tuple[
        Tuple[cython.double, cython.double], Tuple[cython.double, cython.double], Tuple[cython.double, cython.double]
    ],
    pixel_size: cython.double,
) -> vector[Tuple[cython.double, cython.double]]:
    """Convert line segments to a list of pixel centers they intersect with."""

    line_segments: vector[Tuple[Tuple[cython.double, cython.double], Tuple[cython.double, cython.double]]]
    line_segments.push_back((triangle_2d[0], triangle_2d[1]))
    line_segments.push_back((triangle_2d[1], triangle_2d[2]))
    line_segments.push_back((triangle_2d[2], triangle_2d[0]))

    return _line_segments_to_pixels(line_segments, pixel_size)


@cython.cfunc
@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _triangles_to_voxel_matrix(
    triangles: vector[
        Tuple[
            Tuple[cython.double, cython.double, cython.double],
            Tuple[cython.double, cython.double, cython.double],
            Tuple[cython.double, cython.double, cython.double],
        ]
    ],
    n_triangles: cython.int,
    voxel_size: cython.double,
    matrix: bool_C[:, :, :],
    matrix_origin_center: Tuple[cython.double, cython.double, cython.double],
) -> bool_C[:, :, :]:
    """Convert 3D triangles to a voxel matrix representation using intersection tests."""
    # pylint: disable=too-many-locals, too-many-branches, too-many-statements, invalid-name

    # Check interface voxels
    for i in range(n_triangles):
        x_abscissa = _round_to_digits(triangles[i][0][0], 9)
        y_abscissa = _round_to_digits(triangles[i][0][1], 9)
        z_abscissa = _round_to_digits(triangles[i][0][2], 9)

        # Check if two points of the triangle are equal
        if _check_triangle_equal_point(triangles[i]):
            pass

        # Check if the triangle is in the Y-Z plane at the interface between voxels
        elif (
            _round_to_digits(triangles[i][0][0], 9)
            == _round_to_digits(triangles[i][1][0], 9)
            == _round_to_digits(triangles[i][2][0], 9)
        ) and _is_integer(_round_to_digits(x_abscissa / voxel_size, 9)):
            # Define the 3D triangle in 2D
            p0: Tuple[cython.double, cython.double] = (triangles[i][0][1], triangles[i][0][2])
            p1: Tuple[cython.double, cython.double] = (triangles[i][1][1], triangles[i][1][2])
            p2: Tuple[cython.double, cython.double] = (triangles[i][2][1], triangles[i][2][2])

            triangle_2d: Tuple[
                Tuple[cython.double, cython.double],
                Tuple[cython.double, cython.double],
                Tuple[cython.double, cython.double],
            ] = (p0, p1, p2)

            # Compute intersecting pixels
            pixels = _triangle_2d_to_pixels(triangle_2d, voxel_size)

            min_center: Tuple[cython.double, cython.double] = _get_min_pixel_grid_center(pixels)
            max_center: Tuple[cython.double, cython.double] = _get_max_pixel_grid_center(pixels)

            dim_x = cython.cast(cython.int, math_c.round((max_center[0] - min_center[0]) / voxel_size + 1))
            dim_y = cython.cast(cython.int, math_c.round((max_center[1] - min_center[1]) / voxel_size + 1))

            pixel_matrix = _pixel_centers_to_filled_pixel_matrix(pixels, voxel_size, (dim_x, dim_y), min_center)

            # Put the corresponding voxels at True, using the indices
            ix1 = cython.cast(
                cython.int,
                _round_to_digits((x_abscissa - matrix_origin_center[0] - voxel_size / 2) / voxel_size, 9),
            )
            ix2 = cython.cast(
                cython.int,
                _round_to_digits((x_abscissa - matrix_origin_center[0] + voxel_size / 2) / voxel_size, 9),
            )

            dx: cython.int
            dy: cython.int
            for dx in range(dim_x):
                for dy in range(dim_y):
                    if pixel_matrix[dx + 1, dy + 1]:
                        iy = cython.cast(
                            cython.int,
                            _round_to_digits(
                                ((min_center[0] + dx * voxel_size) - matrix_origin_center[1]) / voxel_size, 6
                            ),
                        )
                        iz = cython.cast(
                            cython.int,
                            _round_to_digits(
                                ((min_center[1] + dy * voxel_size) - matrix_origin_center[2]) / voxel_size, 6
                            ),
                        )

                        matrix[ix1, iy, iz] = True
                        matrix[ix2, iy, iz] = True

        # Check if the triangle is in the X-Z plane at the interface between voxels
        elif (
            _round_to_digits(triangles[i][0][1], 9)
            == _round_to_digits(triangles[i][1][1], 9)
            == _round_to_digits(triangles[i][2][1], 9)
        ) and _is_integer(_round_to_digits(y_abscissa / voxel_size, 9)):
            # Define the 3D triangle in 2D
            p0: Tuple[cython.double, cython.double] = (triangles[i][0][0], triangles[i][0][2])
            p1: Tuple[cython.double, cython.double] = (triangles[i][1][0], triangles[i][1][2])
            p2: Tuple[cython.double, cython.double] = (triangles[i][2][0], triangles[i][2][2])

            triangle_2d: Tuple[
                Tuple[cython.double, cython.double],
                Tuple[cython.double, cython.double],
                Tuple[cython.double, cython.double],
            ] = (p0, p1, p2)

            # Compute intersecting pixels
            pixels = _triangle_2d_to_pixels(triangle_2d, voxel_size)

            min_center: Tuple[cython.double, cython.double] = _get_min_pixel_grid_center(pixels)
            max_center: Tuple[cython.double, cython.double] = _get_max_pixel_grid_center(pixels)

            dim_x = cython.cast(cython.int, math_c.round((max_center[0] - min_center[0]) / voxel_size + 1))
            dim_y = cython.cast(cython.int, math_c.round((max_center[1] - min_center[1]) / voxel_size + 1))

            pixel_matrix = _pixel_centers_to_filled_pixel_matrix(pixels, voxel_size, (dim_x, dim_y), min_center)

            # Put the corresponding voxels at True, using the indices
            iy1 = cython.cast(
                cython.int,
                _round_to_digits((y_abscissa - matrix_origin_center[1] - voxel_size / 2) / voxel_size, 9),
            )
            iy2 = cython.cast(
                cython.int,
                _round_to_digits((y_abscissa - matrix_origin_center[1] + voxel_size / 2) / voxel_size, 9),
            )

            dx: cython.int
            dy: cython.int
            for dx in range(dim_x):
                for dy in range(dim_y):
                    if pixel_matrix[dx + 1, dy + 1]:
                        ix = cython.cast(
                            cython.int,
                            _round_to_digits(
                                ((min_center[0] + dx * voxel_size) - matrix_origin_center[0]) / voxel_size, 6
                            ),
                        )
                        iz = cython.cast(
                            cython.int,
                            _round_to_digits(
                                ((min_center[1] + dy * voxel_size) - matrix_origin_center[2]) / voxel_size, 6
                            ),
                        )

                        matrix[ix, iy1, iz] = True
                        matrix[ix, iy2, iz] = True

        # Check if the triangle is in the X-Y plane at the interface between voxels
        elif (
            _round_to_digits(triangles[i][0][2], 9)
            == _round_to_digits(triangles[i][1][2], 9)
            == _round_to_digits(triangles[i][2][2], 9)
        ) and _is_integer(_round_to_digits(z_abscissa / voxel_size, 9)):
            # Define the 3D triangle in 2D
            p0: Tuple[cython.double, cython.double] = (triangles[i][0][0], triangles[i][0][1])
            p1: Tuple[cython.double, cython.double] = (triangles[i][1][0], triangles[i][1][1])
            p2: Tuple[cython.double, cython.double] = (triangles[i][2][0], triangles[i][2][1])

            triangle_2d: Tuple[
                Tuple[cython.double, cython.double],
                Tuple[cython.double, cython.double],
                Tuple[cython.double, cython.double],
            ] = (p0, p1, p2)

            # Compute intersecting pixels
            pixels = _triangle_2d_to_pixels(triangle_2d, voxel_size)

            min_center: Tuple[cython.double, cython.double] = _get_min_pixel_grid_center(pixels)
            max_center: Tuple[cython.double, cython.double] = _get_max_pixel_grid_center(pixels)

            dim_x = cython.cast(cython.int, math_c.round((max_center[0] - min_center[0]) / voxel_size + 1))
            dim_y = cython.cast(cython.int, math_c.round((max_center[1] - min_center[1]) / voxel_size + 1))

            pixel_matrix = _pixel_centers_to_filled_pixel_matrix(pixels, voxel_size, (dim_x, dim_y), min_center)

            # Put the corresponding voxels at True, using the indices
            iz1 = cython.cast(
                cython.int,
                _round_to_digits((z_abscissa - matrix_origin_center[2] - voxel_size / 2) / voxel_size, 9),
            )
            iz2 = cython.cast(
                cython.int,
                _round_to_digits((z_abscissa - matrix_origin_center[2] + voxel_size / 2) / voxel_size, 9),
            )

            dx: cython.int
            dy: cython.int
            for dx in range(dim_x):
                for dy in range(dim_y):
                    if pixel_matrix[dx + 1, dy + 1]:
                        ix = cython.cast(
                            cython.int,
                            _round_to_digits(
                                ((min_center[0] + dx * voxel_size) - matrix_origin_center[0]) / voxel_size, 6
                            ),
                        )
                        iy = cython.cast(
                            cython.int,
                            _round_to_digits(
                                ((min_center[1] + dy * voxel_size) - matrix_origin_center[1]) / voxel_size, 6
                            ),
                        )

                        matrix[ix, iy, iz1] = True
                        matrix[ix, iy, iz2] = True

        # Check intersecting voxels
        else:
            triangle: Tuple[
                Tuple[cython.double, cython.double, cython.double],
                Tuple[cython.double, cython.double, cython.double],
                Tuple[cython.double, cython.double, cython.double],
            ] = triangles[i]
            min_point: Tuple[cython.double, cython.double, cython.double]
            max_point: Tuple[cython.double, cython.double, cython.double]
            min_point, max_point = _triangle_min_max_points(triangle)

            voxel_centers = _aabb_intersecting_boxes(min_point, max_point, voxel_size)
            for j in range(voxel_centers.size()):
                ix = cython.cast(
                    cython.int,
                    _round_to_digits((voxel_centers[j][0] - matrix_origin_center[0]) / voxel_size, 9),
                )
                iy = cython.cast(
                    cython.int,
                    _round_to_digits((voxel_centers[j][1] - matrix_origin_center[1]) / voxel_size, 9),
                )
                iz = cython.cast(
                    cython.int,
                    _round_to_digits((voxel_centers[j][2] - matrix_origin_center[2]) / voxel_size, 9),
                )

                if not matrix[ix, iy, iz]:
                    if _triangle_intersects_voxel(
                        triangle,
                        voxel_centers[j],
                        (0.5 * voxel_size, 0.5 * voxel_size, 0.5 * voxel_size),
                    ):
                        matrix[ix, iy, iz] = True

    return matrix


@cython.cfunc
@cython.exceptval(check=False)
def _get_min_pixel_grid_center(
    pixel_centers: vector[Tuple[cython.double, cython.double]]
) -> Tuple[cython.double, cython.double]:
    """Calculate and return the minimum x and y coordinates among a collection of pixel centers."""

    min_x = min_y = math_c.INFINITY
    for i in range(pixel_centers.size()):
        min_x = min(min_x, pixel_centers[i][0])
        min_y = min(min_y, pixel_centers[i][1])

    return min_x, min_y


@cython.cfunc
@cython.exceptval(check=False)
def _get_max_pixel_grid_center(
    pixel_centers: vector[Tuple[cython.double, cython.double]]
) -> Tuple[cython.double, cython.double]:
    """Calculate and return the maximum x and y coordinates among a collection of pixel centers."""

    max_x = max_y = -math_c.INFINITY
    for i in range(pixel_centers.size()):
        max_x = max(max_x, pixel_centers[i][0])
        max_y = max(max_y, pixel_centers[i][1])

    return max_x, max_y


@cython.cfunc
@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _pixel_centers_to_filled_pixel_matrix(
    pixel_centers: vector[Tuple[cython.double, cython.double]],
    pixel_size: cython.double,
    shape: Tuple[cython.int, cython.int],
    min_center: Tuple[cython.double, cython.double],
) -> bool_C[:, :]:
    """
    Convert pixel centers to a filled boolean matrix.
    """
    # pylint: disable=invalid-name

    matrix: bool_C[:, :] = np.zeros((shape[0] + 2, shape[1] + 2), dtype=np.bool_)

    for i in range(pixel_centers.size()):
        ix = cython.cast(cython.int, _round_to_digits((pixel_centers[i][0] - min_center[0]) / pixel_size, 9)) + 1
        iy = cython.cast(cython.int, _round_to_digits((pixel_centers[i][1] - min_center[1]) / pixel_size, 9)) + 1
        matrix[ix, iy] = True

    matrix_outer_filled = _flood_fill_matrix_2d(matrix.copy(), (0, 0), True, (shape[0] + 2, shape[1] + 2))

    for ix in range(shape[0] + 2):
        for iy in range(shape[1] + 2):
            if not matrix_outer_filled[ix, iy]:
                matrix[ix, iy] = True

    return matrix


@cython.cfunc
@cython.exceptval(check=False)
def _check_triangle_equal_point(
    triangle: Tuple[
        Tuple[cython.double, cython.double, cython.double],
        Tuple[cython.double, cython.double, cython.double],
        Tuple[cython.double, cython.double, cython.double],
    ]
) -> bool_C:
    """Check if any two points of a triangle are equal, implying degeneracy."""

    return (
        (triangle[0][0] == triangle[1][0] and triangle[0][1] == triangle[1][1] and triangle[0][2] == triangle[1][2])
        or (triangle[0][0] == triangle[2][0] and triangle[0][1] == triangle[2][1] and triangle[0][2] == triangle[2][2])
        or (triangle[1][0] == triangle[2][0] and triangle[1][1] == triangle[2][1] and triangle[1][2] == triangle[2][2])
    )


@cython.cfunc
@cython.exceptval(check=False)
def _triangle_min_max_points(
    triangle: Tuple[
        Tuple[cython.double, cython.double, cython.double],
        Tuple[cython.double, cython.double, cython.double],
        Tuple[cython.double, cython.double, cython.double],
    ]
) -> Tuple[Tuple[cython.double, cython.double, cython.double], Tuple[cython.double, cython.double, cython.double]]:
    """Calculate and return the minimum and maximum coordinates of a 3D triangle."""

    min_x: cython.double = math_c.INFINITY
    min_y: cython.double = math_c.INFINITY
    min_z: cython.double = math_c.INFINITY
    max_x: cython.double = -math_c.INFINITY
    max_y: cython.double = -math_c.INFINITY
    max_z: cython.double = -math_c.INFINITY

    min_x = min(min_x, triangle[0][0])
    min_x = min(min_x, triangle[1][0])
    min_x = min(min_x, triangle[2][0])

    min_y = min(min_y, triangle[0][1])
    min_y = min(min_y, triangle[1][1])
    min_y = min(min_y, triangle[2][1])

    min_z = min(min_z, triangle[0][2])
    min_z = min(min_z, triangle[1][2])
    min_z = min(min_z, triangle[2][2])

    max_x = max(max_x, triangle[0][0])
    max_x = max(max_x, triangle[1][0])
    max_x = max(max_x, triangle[2][0])

    max_y = max(max_y, triangle[0][1])
    max_y = max(max_y, triangle[1][1])
    max_y = max(max_y, triangle[2][1])

    max_z = max(max_z, triangle[0][2])
    max_z = max(max_z, triangle[1][2])
    max_z = max(max_z, triangle[2][2])

    return (min_x, min_y, min_z), (max_x, max_y, max_z)


@cython.cfunc
def _triangles_min_max_points(
    triangles: vector[
        Tuple[
            Tuple[cython.double, cython.double, cython.double],
            Tuple[cython.double, cython.double, cython.double],
            Tuple[cython.double, cython.double, cython.double],
        ]
    ]
) -> Tuple[Tuple[cython.double, cython.double, cython.double], Tuple[cython.double, cython.double, cython.double]]:
    """Calculate and return the minimum and maximum coordinates across a collection of 3D triangles."""

    min_x: cython.double = math_c.INFINITY
    min_y: cython.double = math_c.INFINITY
    min_z: cython.double = math_c.INFINITY
    max_x: cython.double = -math_c.INFINITY
    max_y: cython.double = -math_c.INFINITY
    max_z: cython.double = -math_c.INFINITY

    for i in range(triangles.size()):
        min_x = min(min_x, triangles[i][0][0])
        min_x = min(min_x, triangles[i][1][0])
        min_x = min(min_x, triangles[i][2][0])

        min_y = min(min_y, triangles[i][0][1])
        min_y = min(min_y, triangles[i][1][1])
        min_y = min(min_y, triangles[i][2][1])

        min_z = min(min_z, triangles[i][0][2])
        min_z = min(min_z, triangles[i][1][2])
        min_z = min(min_z, triangles[i][2][2])

        max_x = max(max_x, triangles[i][0][0])
        max_x = max(max_x, triangles[i][1][0])
        max_x = max(max_x, triangles[i][2][0])

        max_y = max(max_y, triangles[i][0][1])
        max_y = max(max_y, triangles[i][1][1])
        max_y = max(max_y, triangles[i][2][1])

        max_z = max(max_z, triangles[i][0][2])
        max_z = max(max_z, triangles[i][1][2])
        max_z = max(max_z, triangles[i][2][2])

    return (min_x, min_y, min_z), (max_x, max_y, max_z)


@cython.boundscheck(False)
@cython.wraparound(False)
def _voxel_triangular_faces(
    x: cython.double, y: cython.double, z: cython.double, voxel_size: float
) -> vector[
    Tuple[
        Tuple[cython.double, cython.double, cython.double],
        Tuple[cython.double, cython.double, cython.double],
        Tuple[cython.double, cython.double, cython.double],
    ]
]:
    """Helper method to compute the 12 triangular faces that compose a voxel, for visualization."""
    # pylint: disable=invalid-name

    sx, sy, sz = voxel_size, voxel_size, voxel_size
    hx, hy, hz = sx / 2, sy / 2, sz / 2

    faces: vector[
        Tuple[
            Tuple[cython.double, cython.double, cython.double],
            Tuple[cython.double, cython.double, cython.double],
            Tuple[cython.double, cython.double, cython.double],
        ]
    ]

    # Front face
    faces.push_back(
        (
            _round_point_3d_to_digits((x - hx, y - hy, z + hz), 9),
            _round_point_3d_to_digits((x - hx, y + hy, z + hz), 9),
            _round_point_3d_to_digits((x + hx, y + hy, z + hz), 9),
        )
    )
    faces.push_back(
        (
            _round_point_3d_to_digits((x - hx, y - hy, z + hz), 9),
            _round_point_3d_to_digits((x + hx, y + hy, z + hz), 9),
            _round_point_3d_to_digits((x + hx, y - hy, z + hz), 9),
        )
    )

    # Back face
    faces.push_back(
        (
            _round_point_3d_to_digits((x - hx, y - hy, z - hz), 9),
            _round_point_3d_to_digits((x - hx, y + hy, z - hz), 9),
            _round_point_3d_to_digits((x + hx, y + hy, z - hz), 9),
        )
    )
    faces.push_back(
        (
            _round_point_3d_to_digits((x - hx, y - hy, z - hz), 9),
            _round_point_3d_to_digits((x + hx, y + hy, z - hz), 9),
            _round_point_3d_to_digits((x + hx, y - hy, z - hz), 9),
        )
    )

    # Left face
    faces.push_back(
        (
            _round_point_3d_to_digits((x - hx, y - hy, z - hz), 9),
            _round_point_3d_to_digits((x - hx, y + hy, z - hz), 9),
            _round_point_3d_to_digits((x - hx, y + hy, z + hz), 9),
        )
    )
    faces.push_back(
        (
            _round_point_3d_to_digits((x - hx, y - hy, z - hz), 9),
            _round_point_3d_to_digits((x - hx, y + hy, z + hz), 9),
            _round_point_3d_to_digits((x - hx, y - hy, z + hz), 9),
        )
    )

    # Right face
    faces.push_back(
        (
            _round_point_3d_to_digits((x + hx, y - hy, z - hz), 9),
            _round_point_3d_to_digits((x + hx, y + hy, z - hz), 9),
            _round_point_3d_to_digits((x + hx, y + hy, z + hz), 9),
        )
    )
    faces.push_back(
        (
            _round_point_3d_to_digits((x + hx, y - hy, z - hz), 9),
            _round_point_3d_to_digits((x + hx, y + hy, z + hz), 9),
            _round_point_3d_to_digits((x + hx, y - hy, z + hz), 9),
        )
    )

    # Top face
    faces.push_back(
        (
            _round_point_3d_to_digits((x - hx, y + hy, z - hz), 9),
            _round_point_3d_to_digits((x + hx, y + hy, z - hz), 9),
            _round_point_3d_to_digits((x + hx, y + hy, z + hz), 9),
        )
    )
    faces.push_back(
        (
            _round_point_3d_to_digits((x - hx, y + hy, z - hz), 9),
            _round_point_3d_to_digits((x + hx, y + hy, z + hz), 9),
            _round_point_3d_to_digits((x - hx, y + hy, z + hz), 9),
        )
    )

    # Bottom face
    faces.push_back(
        (
            _round_point_3d_to_digits((x - hx, y - hy, z - hz), 9),
            _round_point_3d_to_digits((x + hx, y - hy, z - hz), 9),
            _round_point_3d_to_digits((x + hx, y - hy, z + hz), 9),
        )
    )
    faces.push_back(
        (
            _round_point_3d_to_digits((x - hx, y - hy, z - hz), 9),
            _round_point_3d_to_digits((x + hx, y - hy, z + hz), 9),
            _round_point_3d_to_digits((x - hx, y - hy, z + hz), 9),
        )
    )

    return faces



"""
Concatenate common operation for two or more objects.

"""
import math
import random

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import least_squares
import scipy.integrate as scipy_integrate
from sklearn.cluster import DBSCAN

import volmdlr.core
from volmdlr.core import EdgeStyle


def plot_circle(circle, ax=None, edge_style: EdgeStyle = EdgeStyle()):
    """
    Create a Matplotlib plot for a circle 2d or fullarc 2d.

    :param circle: circle to plot.
    :param ax: Matplotlib plot axis.
    :param edge_style: Edge Style to implement.
    :return: Matplotlib plot axis.
    """
    if ax is None:
        _, ax = plt.subplots()
    if circle.radius > 0:
        ax.add_patch(matplotlib.patches.Arc((circle.center.x, circle.center.y),
                                            2 * circle.radius,
                                            2 * circle.radius,
                                            angle=0,
                                            theta1=0,
                                            theta2=360,
                                            color=edge_style.color,
                                            alpha=edge_style.alpha,
                                            linestyle=edge_style.linestyle,
                                            linewidth=edge_style.linewidth))
    if edge_style.plot_points:
        ax.plot([circle.start.x], [circle.start.y], 'o',
                color=edge_style.color, alpha=edge_style.alpha)
    if edge_style.equal_aspect:
        ax.set_aspect('equal')
    x_min, x_max = circle.center[0] - circle.radius, circle.center[0] + circle.radius
    y_min, y_max = circle.center[1] - circle.radius, circle.center[1] + circle.radius
    ax.set_xlim(x_min, x_max)
    ax.set_ylim(y_min, y_max)

    return ax


def random_color():
    """Random color generator."""
    return random.random(), random.random(), random.random()


def split_wire_by_plane(wire, plane3d):
    """
    Splits a wire into two parts using a plane.

    This method splits a wire into two parts based on the intersection points between the wire's primitives
    (edges) and a given 3D plane. It first finds the intersection points between each primitive and the plane,
    excluding duplicate points. Then, it checks if the number of intersection points is greater than one. If so,
    it raises a NotImplementedError, as the split is ambiguous. Otherwise, it performs the split using the
    `split_with_sorted_points` method of the wire object. The resulting wire objects are returned as a tuple.
    Note: The method assumes that the wire and the plane are in the same coordinate system.

    :param wire: The wire object to be split.
    :param plane3d: The 3D plane object used for splitting the wire.
    :return: A tuple containing two wire objects resulting from the split.
    :raises: NotImplementedError: If the wire intersects the plane at more than one point.

    :Example:
    >>> from volmdlr import Point3D, Vector3D
    >>> from volmdlr.surfaces import Plane3D
    >>> from volmdlr.core import EdgeStyle
    >>> from volmdlr.utils.common_operations import random_color
    >>> from volmdlr.models.open_rounded_line_segments import open_rounded_line_segements
    >>> plane = Plane3D.from_plane_vectors(Point3D(0.4, 0.4, 0.2), Vector3D(1, 0, 0), Vector3D(0, 1, 0))
    >>> split_wire1,split_wire2 = split_wire_by_plane(open_rounded_line_segements, plane)
    >>> ax = open_rounded_line_segements.plot()
    >>> plane.plot(ax)
    >>> split_wire1.plot(ax, EdgeStyle(random_color()))
    >>> split_wire2.plot(ax, EdgeStyle(random_color()))
    """
    wire_plane_intersections = []
    for primitive in wire.primitives:
        intersections = plane3d.edge_intersections(primitive)
        for intersection in intersections:
            if not volmdlr.core.point_in_list(intersection, wire_plane_intersections):
                wire_plane_intersections.append(intersection)
    if len(wire_plane_intersections) > 1:
        raise NotImplementedError
    wire1, wire2 = wire.split_with_sorted_points([wire_plane_intersections[0], wire.primitives[-1].end])
    return wire1, wire2


def plot_components_from_points(points, close_plot: bool = False):
    """
    Gets Matplotlib components from points.

    :param points: given points.
    :param close_plot: Weather to close the plot or not.
    :return:
    """
    components = [[], [], []]
    for point in points:
        for i, component in enumerate(point):
            components[i].append(component)
    valid_components = []
    for list_components in components:
        if list_components:
            if close_plot:
                list_components.append(list_components[0])
            valid_components.append(list_components)
    return valid_components


def plot_from_discretization_points(ax, edge_style, element, number_points: int = None, close_plot: bool = False):
    """
    General plot method using discretization_points method to generate points.

    :param ax: Matplotlib plot.
    :param edge_style: edge_style to be applied to plot.
    :param element: volmdlr element to be plotted (either 2D or 3D).
    :param number_points: number of points to be used in the plot.
    :param close_plot: specifies if plot is to be closed or not.
    :return: Matplotlib plot axis.
    """
    points = element.discretization_points(number_points=number_points)
    valid_components = plot_components_from_points(points, close_plot)
    ax.plot(*valid_components, color=edge_style.color, alpha=edge_style.alpha)
    return ax


def minimum_distance_points_circle3d_linesegment3d(circle3d,  linesegment3d):
    """
    Gets the points from the arc and the line that gives the minimal distance between them.

    :param circle3d: Circle 3d or Arc 3d.
    :param linesegment3d: Other line segment 3d.
    :return: Minimum distance points.
    """
    def distance_squared(x, u_param, v_param, k_param, w_param):
        """Calculates the squared distance."""
        return (u_param.dot(u_param) * x[0] ** 2 + w_param.dot(w_param) + v_param.dot(v_param) * (
                (math.sin(x[1])) ** 2) * radius ** 2 + k_param.dot(k_param) * ((math.cos(x[1])) ** 2) * radius ** 2
                - 2 * x[0] * w_param.dot(u_param) - 2 * x[0] * radius * math.sin(x[1]) * u_param.dot(v_param) - 2 * x[
                    0] * radius * math.cos(x[1]) * u_param.dot(k_param)
                + 2 * radius * math.sin(x[1]) * w_param.dot(v_param) +
                2 * radius * math.cos(x[1]) * w_param.dot(k_param)
                + math.sin(2 * x[1]) * v_param.dot(k_param) * radius ** 2)
    circle_point = circle3d.point_at_abscissa(0.0)
    radius = circle3d.radius
    linseg_direction_vector = linesegment3d.direction_vector()
    vector_point_origin = circle_point - circle3d.frame.origin
    vector_point_origin = vector_point_origin.unit_vector()
    w = circle3d.frame.origin - linesegment3d.start
    v = circle3d.frame.w.cross(vector_point_origin)

    results = []
    for initial_value in [np.array([0.5, circle3d.angle / 2]), np.array([0.5, 0]), np.array([0.5, circle3d.angle])]:
        results.append(least_squares(distance_squared, initial_value,
                                     bounds=[(0, 0), (1, circle3d.angle)],
                                     args=(linseg_direction_vector, v, vector_point_origin, w)))

    point1 = linesegment3d.point_at_abscissa(results[0].x[0] * linesegment3d.length())
    point2 = circle3d.point_at_abscissa(results[1].x[1] * circle3d.radius)

    for couple in results[1:]:
        ptest1 = linesegment3d.point_at_abscissa(couple.x[0] * linesegment3d.length())
        ptest2 = circle3d.point_at_abscissa(couple.x[1] * circle3d.radius)
        dtest = ptest1.point_distance(ptest2)
        if dtest < v.dot(v):
            point1, point2 = ptest1, ptest2

    return point1, point2


def get_abscissa_discretization(primitive, abscissa1, abscissa2, max_number_points: int = 10,
                                return_abscissas: bool = True):
    """
    Gets n discretization points between two given points of the edge.

    :param primitive: Primitive to discretize locally.
    :param abscissa1: Initial abscissa.
    :param abscissa2: Final abscissa.
    :param max_number_points: Expected number of points to discretize locally.
    :param return_abscissas: By default, returns also a list of abscissas corresponding to the
        discretization points
    :return: list of locally discretized point and a list containing the abscissas' values.
    """
    discretized_points_between_1_2 = []
    points_abscissas = []
    for abscissa in np.linspace(abscissa1, abscissa2, num=max_number_points):
        if abscissa > primitive.length() + 1e-6:
            continue
        abscissa_point = primitive.point_at_abscissa(abscissa)
        if not volmdlr.core.point_in_list(abscissa_point, discretized_points_between_1_2):
            discretized_points_between_1_2.append(abscissa_point)
            points_abscissas.append(abscissa)
    if return_abscissas:
        return discretized_points_between_1_2, points_abscissas
    return discretized_points_between_1_2


def get_point_distance_to_edge(edge, point, start, end):
    """
    Calculates the distance from a given point to an edge.

    :param edge: Edge to calculate distance to point.
    :param point: Point to calculate the distance to edge.
    :param start: Edge's start point.
    :param end: Edge's end point.
    :return: distance to edge.
    """
    best_distance = math.inf
    abscissa1 = 0
    abscissa2 = edge.length()
    distance = best_distance
    point1_ = start
    point2_ = end
    number_points = 10 if abs(abscissa2 - abscissa1) > 5e-6 else 2
    linesegment_class_ = getattr(volmdlr.edges, 'LineSegment' + edge.__class__.__name__[-2:])
    while True:
        discretized_points_between_1_2 = edge.local_discretization(point1_, point2_, number_points)
        if not discretized_points_between_1_2:
            break
        distance = point.point_distance(discretized_points_between_1_2[0])
        for point1, point2 in zip(discretized_points_between_1_2[:-1], discretized_points_between_1_2[1:]):
            if point1.is_close(point2):
                continue
            line = linesegment_class_(point1, point2)
            dist = line.point_distance(point)
            if dist < distance:
                point1_ = point1
                point2_ = point2
                distance = dist
        if not point1_ or math.isclose(distance, best_distance, abs_tol=1e-7):
            break
        best_distance = distance
        if math.isclose(abscissa1, abscissa2, abs_tol=1e-6):
            break
    return distance


def ellipse_abscissa_angle_integration(ellipse3d, point_abscissa, angle_start, initial_angle):
    """
    Calculates the angle for a given abscissa point by integrating the ellipse.

    :param ellipse3d: the Ellipse3D.
    :param point_abscissa: the given abscissa for given point.
    :param angle_start: Ellipse3D / ArcEllipse3D start angle. (0 for Ellipse3D).
    :param initial_angle: angle abscissa's initial value.
    :return: final angle abscissa's value.
    """
    def ellipse_arc_length(theta):
        return math.sqrt((ellipse3d.major_axis ** 2) * math.sin(theta) ** 2 +
                         (ellipse3d.minor_axis ** 2) * math.cos(theta) ** 2)

    iter_counter = 0
    while True:
        res, _ = scipy_integrate.quad(ellipse_arc_length, angle_start, initial_angle)
        if math.isclose(res, point_abscissa, abs_tol=1e-8):
            abscissa_angle = initial_angle
            break
        if res > point_abscissa:
            increment_factor = (abs(initial_angle - angle_start) * (point_abscissa - res)) / (2 * abs(res))
        elif res == 0.0:
            increment_factor = 1e-5
        else:
            increment_factor = (abs(initial_angle - angle_start) * (point_abscissa - res)) / abs(res)
        initial_angle += increment_factor
        iter_counter += 1
    return abscissa_angle


def get_plane_equation_coefficients(plane_frame):
    """
    Returns the a,b,c,d coefficient from equation ax+by+cz+d = 0.

    """
    a, b, c = plane_frame.w
    d = -plane_frame.origin.dot(plane_frame.w)
    return round(a, 12), round(b, 12), round(c, 12), round(d, 12)


def order_points_list_for_nearest_neighbor(points):
    """
    Given a list of unordered points defining a path, it will order these points considering the nearest neighbor.

    """
    ordered_points = []
    remaining_points = points[:]
    current_point = remaining_points.pop(0)

    while remaining_points:
        nearest_point_idx = np.argmin([current_point.point_distance(p)for p in remaining_points])
        nearest_point = remaining_points.pop(nearest_point_idx)
        ordered_points.append(current_point)
        current_point = nearest_point

    # Add the last point to complete the loop
    ordered_points.append(current_point)

    return ordered_points


def separate_points_by_closeness(points):
    """
    Separates a list of 3D Cartesian points into two groups based on their spatial closeness using DBSCAN.

    This function applies the DBSCAN (Density-Based Spatial Clustering of Applications with Noise) algorithm to
    the given list of 3D Cartesian points. DBSCAN clusters the points based on their spatial proximity.
    The points are separated into two groups, 'group1' and 'group2', depending on their spatial closeness
    as determined by the DBSCAN clustering.

    Please note that the 'eps' parameter inside the function can be adjusted to control the closeness threshold.

    :param points: A list of 3D Cartesian points, where each point is represented as a list of three coordinates.

    :return:
    - group1 (list of lists): The first group of points based on their closeness.
    - group2 (list of lists): The second group of points based on their closeness.
    """
    points_ = np.array([[point[0], point[1], point[2]] for point in points])

    # Apply DBSCAN clustering with a small epsilon to separate close points
    eps = 0.5
    dbscan = DBSCAN(eps=eps, min_samples=1)
    labels = dbscan.fit_predict(points_)

    # Initialize two empty lists for the two groups
    group1, group2 = [], []

    # Assign points to group1 or group2 based on DBSCAN labels
    for i, label in enumerate(labels):
        if label == 0:
            group1.append(points[i])
        else:
            group2.append(points[i])
    group1 = order_points_list_for_nearest_neighbor(group1)
    group1.append(group1[0])
    if not group2:
        return [group1]
    group2 = order_points_list_for_nearest_neighbor(group2)
    group2.append(group2[0])
    return [group1, group2]



"""
Nurbs main operations algorithms.
"""

from functools import lru_cache
import numpy as np

import volmdlr
from volmdlr.nurbs import core


def knot_insertion(degree, knotvector, ctrlpts, u, **kwargs):
    """
    Computes the control points of the rational/non-rational spline after knot insertion.

    Part of Algorithm A5.1 of The NURBS Book by Piegl & Tiller, 2nd Edition.

    Keyword Arguments:
        * ``num``: number of knot insertions. *Default: 1*
        * ``s``: multiplicity of the knot. *Default: computed via :func:`.find_multiplicity`*
        * ``span``: knot span. *Default: computed via :func:`.find_span_linear`*

    :param degree: degree
    :type degree: int
    :param knotvector: knot vector
    :type knotvector: list, tuple
    :param ctrlpts: control points
    :type ctrlpts: list
    :param u: knot to be inserted
    :type u: float
    :return: updated control points
    :rtype: list

    """
    # Get keyword arguments
    num_insertions = kwargs.get('num', 1)  # number of knot insertions
    knot_multiplicity = kwargs.get('s', core.find_multiplicity(u, knotvector))  # multiplicity
    knot_span = kwargs.get('span', core.find_span_linear(degree, knotvector, len(ctrlpts), u))  # knot span

    # Initialize variables
    num_ctrlpts_new = len(ctrlpts) + num_insertions

    # Initialize new control points array (control points may be weighted or not)
    ctrlpts_new = [np.empty_like(ctrlpts[0]) for _ in range(num_ctrlpts_new)]

    # Initialize a local array of length p + 1
    temp = [np.empty_like(ctrlpts[0]) for _ in range(degree + 1)]

    # Save unaltered control points
    ctrlpts_new[:knot_span - degree + 1] = ctrlpts[:knot_span - degree + 1]
    ctrlpts_new[knot_span - knot_multiplicity + num_insertions:] = ctrlpts[knot_span - knot_multiplicity:]

    # Start filling the temporary local array which will be used to update control points during knot insertion
    for i in range(degree - knot_multiplicity + 1):
        temp[i] = np.array(ctrlpts[knot_span - degree + i])

    # Insert knot "num_insertions" times
    for j in range(1, num_insertions + 1):
        new_knot_index = knot_span - degree + j
        for i in range(degree - j - knot_multiplicity + 1):
            alpha = knot_insertion_alpha(u, tuple(knotvector), knot_span, i, new_knot_index)
            temp[i] = alpha * temp[i + 1] + (1.0 - alpha) * temp[i]
        ctrlpts_new[new_knot_index] = temp[0]
        ctrlpts_new[knot_span + num_insertions - j - knot_multiplicity] = \
            temp[degree - j - knot_multiplicity]

    # Load remaining control points
    new_knot_index = knot_span - degree + num_insertions
    ctrlpts_new[new_knot_index + 1:knot_span - knot_multiplicity] = \
        temp[1:knot_span - knot_multiplicity - new_knot_index + 1]

    # Return control points after knot insertion
    return ctrlpts_new


@lru_cache(maxsize=128)
def knot_insertion_alpha(u, knotvector, span, idx, leg):
    """
    Computes :math:`\\alpha` coefficient for knot insertion algorithm.

    :param u: knot
    :type u: float
    :param knotvector: knot vector
    :type knotvector: tuple
    :param span: knot span
    :type span: int
    :param idx: index value (degree-dependent)
    :type idx: int
    :param leg: i-th leg of the control points polygon
    :type leg: int
    :return: coefficient value
    :rtype: float
    """
    return (u - knotvector[leg + idx]) / (knotvector[idx + span + 1] - knotvector[leg + idx])


def knot_insertion_kv(knotvector, u, span, num_insertions):
    """
    Computes the knot vector of the rational/non-rational spline after knot insertion.

    Part of Algorithm A5.1 of The NURBS Book by Piegl & Tiller, 2nd Edition.

    :param knotvector: knot vector
    :type knotvector: list, tuple
    :param u: knot
    :type u: float
    :param span: knot span
    :type span: int
    :param num_insertions: number of knot insertions
    :type num_insertions: int
    :return: updated knot vector
    :rtype: list
    """
    kv_size = len(knotvector)
    kv_updated = [0.0 for _ in range(kv_size + num_insertions)]

    # Compute new knot vector
    for i in range(0, span + 1):
        kv_updated[i] = knotvector[i]
    for i in range(1, num_insertions + 1):
        kv_updated[span + i] = u
    for i in range(span + 1, kv_size):
        kv_updated[i + num_insertions] = knotvector[i]

    # Return the new knot vector
    return kv_updated


def insert_knot_curve(obj, param, num, **kwargs):
    """
    Inserts knots n-times to a spline geometry.

    Keyword Arguments:
        * ``check_num``: enables/disables operation validity checks. *Default: True*

    :param obj: spline geometry
    :param param: knot(s) to be inserted in [u, v, w] format
    :type param: list, tuple
    :param num: number of knot insertions in [num_u, num_v, num_w] format
    :type num: list, tuple
    :return: updated spline geometry

    """
    # Get keyword arguments
    check_num = kwargs.get('check_num', True)  # can be set to False when the caller checks number of insertions

    if check_num:
        # Check the validity of number of insertions
        if not isinstance(num, (list, tuple)):
            raise ValueError("The number of insertions must be a list or a tuple")

        if len(num) != 1:
            raise ValueError("The length of the num array must be equal to the number of parametric dimensions")

        for val in num:
            if val < 0:
                raise ValueError('Number of insertions must be a positive integer value')

    if param[0] is not None and num[0] > 0:
        # Find knot multiplicity
        param_multiplicity = core.find_multiplicity(param[0], obj.knotvector)

        # Check if it is possible add that many number of knots
        if check_num and num[0] > obj.degree - param_multiplicity:
            raise ValueError("Knot " + str(param[0]) + " cannot be inserted " + str(num[0]) + " times")

        # Find knot span
        span = core.find_span_linear(obj.degree, obj.knotvector, len(obj.ctrlpts), param[0])

        # Compute new knot vector
        kv_new = knot_insertion_kv(obj.knotvector, param[0], span, num[0])

        # Compute new control points
        cpts = obj.ctrlptsw if obj.rational else obj.ctrlpts
        cpts_tmp = knot_insertion(obj.degree, obj.knotvector, cpts, param[0],
                                  num=num[0], s=param_multiplicity, span=span)
        weights = None
        if obj.rational:
            cpts_tmp, weights = separate_ctrlpts_weights(cpts_tmp)

        # Create new curve
        knots, knot_multiplicities = get_knots_and_multiplicities(kv_new)
        point_name = "Point" + obj.__class__.__name__[-2:]
        cpts_tmp = [getattr(volmdlr, point_name)(*point) for point in cpts_tmp]
    # Return new spline geometry
    return obj.__class__(obj.degree, cpts_tmp, knot_multiplicities, knots, weights)


def split_curve(obj, param, **kwargs):
    """
    Splits the curve at the input parametric coordinate.

    This method splits the curve into two pieces at the given parametric coordinate, generates two different
    curve objects and returns them. It does not modify the input curve.

    Keyword Arguments:
        * ``find_span_func``: FindSpan implementation. *Default:* :func:`.helpers.find_span_linear`
        * ``insert_knot_func``: knot insertion algorithm implementation. *Default:* :func:`.operations.insert_knot`

    :param obj: Curve to be split
    :type obj: abstract.Curve
    :param param: parameter
    :type param: float
    :return: a list of curve segments
    :rtype: list

    """
    if param in set(obj.domain):
        raise ValueError("Cannot split from the domain edge")

    # Find multiplicity of the knot and define how many times we need to add the knot
    knot_multiplicity = core.find_multiplicity(param, obj.knotvector)
    insertion_count = obj.degree - knot_multiplicity
    knot_span = core.find_span_linear(obj.degree, obj.knotvector, len(obj.ctrlpts), param) - obj.degree + 1

    # Insert knot
    temp_obj = insert_knot_curve(obj, [param], num=[insertion_count], check_num=False)

    # Knot vectors
    curve1_kv, curve2_kv = helper_split_knot_vectors(temp_obj.degree, temp_obj.knotvector, len(temp_obj.ctrlpts),
                                                     param,
                                                     core.find_span_linear)

    return construct_split_curve(temp_obj, curve1_kv, curve2_kv, knot_span, insertion_count)


def construct_split_curve(obj, curve1_kv, curve2_kv, knot_span, insertion_count):
    """
    Helper function to instantiate split curve.
    """
    control_points = obj.control_points
    curve1_ctrlpts = control_points[0:knot_span + insertion_count]
    curve2_ctrlpts = control_points[knot_span + insertion_count - 1:]
    if obj.rational:
        curve1_weights = obj.weights[0:knot_span + insertion_count]
        curve2_weights = obj.weights[knot_span + insertion_count - 1:]
    else:
        curve1_weights = None
        curve2_weights = None

    knots_1, knot_multiplicities_1 = get_knots_and_multiplicities(curve1_kv)

    knots_2, knot_multiplicities_2 = get_knots_and_multiplicities(curve2_kv)

    # Return the split curves
    return [obj.__class__(obj.degree, curve1_ctrlpts, knot_multiplicities_1, knots_1, curve1_weights),
            obj.__class__(obj.degree, curve2_ctrlpts, knot_multiplicities_2, knots_2, curve2_weights)]


def insert_knot_surface(obj, param, num, **kwargs):
    """
    Inserts knots n-times to a spline geometry.

    Keyword Arguments:
        * ``check_num``: enables/disables operation validity checks. *Default: True*

    :param obj: spline geometry
    :param param: knot(s) to be inserted in [u, v, w] format
    :type param: list, tuple
    :param num: number of knot insertions in [num_u, num_v, num_w] format
    :type num: list, tuple
    :return: updated spline geometry

    """
    # Get keyword arguments
    check_num = kwargs.get('check_num', True)  # can be set to False when the caller checks number of insertions
    # u-direction
    if param[0] is not None and num[0] > 0:
        # Find knot multiplicity
        knotvector = obj.knots_vector_u
        param_multiplicity = core.find_multiplicity(param[0], knotvector)

        # Check if it is possible add that many number of knots
        if check_num and num[0] > obj.degree_u - param_multiplicity:
            raise ValueError("Knot " + str(param[0]) + " cannot be inserted " + str(num[0]) + " times (u-dir)")

        # Find knot span
        span = core.find_span_linear(obj.degree_u, knotvector, obj.nb_u, param[0])

        # Compute new knot vector
        new_kv = knot_insertion_kv(knotvector, param[0], span, num[0])

        # Get curves
        cpts_tmp = []
        cpts = obj.ctrlptsw if obj.rational else obj.ctrlpts
        for v in range(obj.nb_v):
            ctrlpts = [cpts[v + (obj.nb_v * u)] for u in range(obj.nb_u)]
            ctrlpts_tmp = knot_insertion(obj.degree_u, knotvector, ctrlpts, param[0],
                                         num=num[0], s=param_multiplicity, span=span)
            cpts_tmp += ctrlpts_tmp

        # Update the surface after knot insertion
        knots, multiplicities = get_knots_and_multiplicities(new_kv)
        cpts_tmp = flip_ctrlpts_u(cpts_tmp, obj.nb_u + num[0], obj.nb_v)
        weights = None
        if obj.rational:
            cpts_tmp, weights = separate_ctrlpts_weights(cpts_tmp)
        control_points = [volmdlr.Point3D(*point) for point in cpts_tmp]
        obj = obj.__class__(obj.degree_u, obj.degree_v,
                            control_points,
                            obj.nb_u + num[0], obj.nb_v, multiplicities, obj.v_multiplicities,
                            knots, obj.v_knots, weights)
    # v-direction
    if param[1] is not None and num[1] > 0:
        # Find knot multiplicity
        knotvector = obj.knots_vector_v
        param_multiplicity = core.find_multiplicity(param[1], knotvector)

        # Check if it is possible add that many number of knots
        if check_num and num[1] > obj.degree_v - param_multiplicity:
            raise ValueError("Knot " + str(param[1]) + " cannot be inserted " + str(num[1]) + " times (v-dir)")

        # Find knot span
        span = core.find_span_linear(obj.degree_v, knotvector, obj.nb_v, param[1])

        # Compute new knot vector
        new_kv = knot_insertion_kv(knotvector, param[1], span, num[1])

        # Get curves
        cpts_tmp = []
        cpts = obj.ctrlptsw if obj.rational else obj.ctrlpts
        for u in range(obj.nb_u):
            ctrlpts = [cpts[v + (obj.nb_v * u)] for v in range(obj.nb_v)]
            ctrlpts_tmp = knot_insertion(obj.degree_v, knotvector, ctrlpts, param[1],
                                         num=num[1], s=param_multiplicity, span=span)
            cpts_tmp += ctrlpts_tmp

        knots, multiplicities = get_knots_and_multiplicities(new_kv)
        weights = None
        if obj.rational:
            cpts_tmp, weights = separate_ctrlpts_weights(cpts_tmp)
        control_points = [volmdlr.Point3D(*point) for point in cpts_tmp]
        # Update the surface after knot insertion
        obj = obj.__class__(obj.degree_u, obj.degree_v,
                            control_points, obj.nb_u, obj.nb_v + num[1], obj.u_multiplicities, multiplicities,
                            obj.u_knots, knots, weights)
    return obj


def split_surface_u(obj, param, **kwargs):
    """
    Splits the surface at the input parametric coordinate on the u-direction.

    This method splits the surface into two pieces at the given parametric coordinate on the u-direction,
    generates two different surface objects and returns them. It does not modify the input surface.

    Keyword Arguments:
        * ``find_span_func``: FindSpan implementation. *Default:* :func:`.helpers.find_span_linear`
        * ``insert_knot_func``: knot insertion algorithm implementation. *Default:* :func:`.operations.insert_knot`

    :param obj: surface
    :type obj: abstract.Surface
    :param param: parameter for the u-direction
    :type param: float
    :return: a list of surface patches
    :rtype: list

    """

    if param in (obj.domain[0], obj.domain[1]):
        raise ValueError("Cannot split from the u-domain edge")

    # Keyword arguments
    span_func = kwargs.get('find_span_func', core.find_span_linear)  # FindSpan implementation
    insert_knot_func = kwargs.get('insert_knot_func', insert_knot_surface)  # Knot insertion algorithm

    # Find multiplicity of the knot
    knotvector_u = obj.knots_vector_u
    knot_span = span_func(obj.degree_u, knotvector_u, obj.nb_u, param) - obj.degree_u + 1
    knot_multiplicity = core.find_multiplicity(param, knotvector_u)
    insertion_count = obj.degree_u - knot_multiplicity

    # Split the original surface
    temp_obj = insert_knot_func(obj, [param, None], num=[insertion_count, 0], check_num=False)

    # Knot vectors
    knotvectors = helper_split_knot_vectors(temp_obj.degree_u, temp_obj.knots_vector_u,
                                            temp_obj.nb_u, param, span_func)

    return construct_split_surfaces(temp_obj, knotvectors, "u", knot_span, insertion_count)


def split_surface_v(obj, param, **kwargs):
    """
    Splits the surface at the input parametric coordinate on the v-direction.

    This method splits the surface into two pieces at the given parametric coordinate on the v-direction,
    generates two different surface objects and returns them. It does not modify the input surface.

    Keyword Arguments:
        * ``find_span_func``: FindSpan implementation. *Default:* :func:`.helpers.find_span_linear`
        * ``insert_knot_func``: knot insertion algorithm implementation. *Default:* :func:`.operations.insert_knot`

    :param obj: surface
    :type obj: abstract.Surface
    :param param: parameter for the v-direction
    :type param: float
    :return: a list of surface patches
    :rtype: list

    """
    if param in (obj.domain[2], obj.domain[3]):
        raise ValueError("Cannot split from the v-domain edge")

    # Keyword arguments
    span_func = kwargs.get('find_span_func', core.find_span_linear)  # FindSpan implementation
    insert_knot_func = kwargs.get('insert_knot_func', insert_knot_surface)  # Knot insertion algorithm

    # Find multiplicity of the knot
    knotvector_v = obj.knots_vector_v
    knot_span = span_func(obj.degree_v, knotvector_v, obj.nb_v, param) - obj.degree_v + 1
    knot_multiplicity = core.find_multiplicity(param, knotvector_v)
    insertion_count = obj.degree_v - knot_multiplicity

    # Split the original surface
    temp_obj = insert_knot_func(obj, [None, param], num=[0, insertion_count], check_num=False)

    # Knot vectors
    knotvectors = helper_split_knot_vectors(temp_obj.degree_v, temp_obj.knots_vector_v, temp_obj.nb_v, param,
                                                   span_func)
    return construct_split_surfaces(temp_obj, knotvectors, "v", knot_span, insertion_count)

def separate_ctrlpts_weights(ctrlptsw):
    """
    Divides weighted control points by weights to generate unweighted control points and weights vector.

    This function is dimension agnostic, i.e. control points can be in any dimension but the last element of the array
    should indicate the weight.

    :param ctrlptsw: weighted control points
    :type ctrlptsw: list, tuple
    :return: unweighted control points and weights vector
    :rtype: list
    """
    ctrlpts = []
    weights = []
    for ptw in ctrlptsw:
        temp = [float(pw / ptw[-1]) for pw in ptw[:-1]]
        ctrlpts.append(temp)
        weights.append(ptw[-1])

    return [ctrlpts, weights]


def flip_ctrlpts_u(ctrlpts, size_u, size_v):
    """
    Flips a list of 1-dimensional control points from u-row order to v-row order.

    **u-row order**: each row corresponds to a list of u values

    **v-row order**: each row corresponds to a list of v values

    :param ctrlpts: control points in u-row order
    :type ctrlpts: list, tuple
    :param size_u: size in u-direction
    :type size_u: int
    :param size_v: size in v-direction
    :type size_v: int
    :return: control points in v-row order
    :rtype: list
    """
    new_ctrlpts = []
    for i in range(0, size_u):
        for j in range(0, size_v):
            temp = [float(c) for c in ctrlpts[i + (j * size_u)]]
            new_ctrlpts.append(temp)

    return new_ctrlpts


def ctrlpts2d_to_ctrlpts(control_points_table):
    """
    Transform the control_points m x n.
    """
    size_u = len(control_points_table)
    size_v = len(control_points_table[0])

    # Make sure that all numbers are float type
    ctrlpts = [[] for _ in range(size_u * size_v)]
    for u in range(size_u):
        for v in range(size_v):
            idx = v + (size_v * u)
            ctrlpts[idx] = [float(coord) for coord in control_points_table[u][v]]

    return ctrlpts


def helper_split_knot_vectors(degree, knotvector, num_ctrlpts, param, span_func):
    """
    Computes knot vectors to split object into two pieces.
    """
    knot_span_new = span_func(degree, knotvector, num_ctrlpts, param) + 1
    kv_1 = list(knotvector[0:knot_span_new])
    kv_1.append(param)
    kv_2 = list(knotvector[knot_span_new:])
    for _ in range(0, degree + 1):
        kv_2.insert(0, param)
    return kv_1, kv_2


def get_knots_and_multiplicities(knotvector):
    """
    Get knots and multiplicities from knotvector in u and v direction.
    """
    knotvector = np.round(knotvector, decimals=17)
    knots = np.unique(knotvector).tolist()
    multiplicities = [core.find_multiplicity(knot, knotvector) for knot in knots]
    return knots, multiplicities


def construct_split_surfaces(obj, knotvectors, direction, knot_span, insertion_count):
    """
    Helper function to construct split surfaces.
    """
    surf1_kv, surf2_kv = knotvectors
    if direction == "u":
        ctrlpts2d = obj.ctrlpts2d()
        surf1_ctrlpts = ctrlpts2d_to_ctrlpts(ctrlpts2d[0:knot_span + insertion_count])
        surf2_ctrlpts = ctrlpts2d_to_ctrlpts(ctrlpts2d[knot_span + insertion_count - 1:])
        u_knots, u_multiplicities = get_knots_and_multiplicities(surf1_kv)
        v_knots, v_multiplicities = get_knots_and_multiplicities(obj.knots_vector_v)
        surf1_nb_u = knot_span + insertion_count
        surf1_nb_v = obj.nb_v
        surf2_nb_u = obj.nb_u - (knot_span + insertion_count - 1)
        surf2_nb_v = obj.nb_v
    else:
        surf1_ctrlpts = []
        surf2_ctrlpts = []
        for v_row in obj.ctrlpts2d():
            temp = v_row[0:knot_span + insertion_count]
            surf1_ctrlpts.extend(temp)
            temp = v_row[knot_span + insertion_count - 1:]
            surf2_ctrlpts.extend(temp)
        u_knots, u_multiplicities = get_knots_and_multiplicities(obj.knots_vector_u)
        v_knots, v_multiplicities = get_knots_and_multiplicities(surf1_kv)
        surf1_nb_u = obj.nb_u
        surf1_nb_v = knot_span + insertion_count
        surf2_nb_u = obj.nb_u
        surf2_nb_v = obj.nb_v - (knot_span + insertion_count - 1)
    weights = None
    if obj.rational:
        surf1_ctrlpts, weights = separate_ctrlpts_weights(surf1_ctrlpts)
    control_points = [volmdlr.Point3D(*point) for point in surf1_ctrlpts]
    surf1 = obj.__class__(obj.degree_u, obj.degree_v, control_points, surf1_nb_u,
                          surf1_nb_v, u_multiplicities, v_multiplicities, u_knots, v_knots, weights)

    # knots
    if direction == "u":
        u_knots, u_multiplicities = get_knots_and_multiplicities(surf2_kv)
    else:
        v_knots, v_multiplicities = get_knots_and_multiplicities(surf2_kv)
    weights = None
    if obj.rational:
        surf2_ctrlpts, weights = separate_ctrlpts_weights(surf2_ctrlpts)
    control_points = [volmdlr.Point3D(*point) for point in surf2_ctrlpts]
    surf2 = obj.__class__(obj.degree_u, obj.degree_v, control_points,
                          surf2_nb_u, surf2_nb_v, u_multiplicities, v_multiplicities, u_knots, v_knots, weights)

    # Return the new surfaces
    return [surf1, surf2]
